# $Id: SpeciesGraph.pm,v 1.23 2007/07/06 04:47:25 faeder Exp $

package SpeciesGraph;

use Class::Struct;
use FindBin;
use lib $FindBin::Bin;
use Molecule;
use Component;
use Compartment;
use BNGUtils;
use Map;
use MoleculeTypesList;
use HNauty;

struct SpeciesGraph => {
  Name=>      '$',			
  Compartment=>'Compartment',
  Molecules=> '@',
  Edges=>     '@',
  Adjacency=> '%',
  StringID=>  '$',
  StringExact=> '$',		
  Quantifier=> '$',
  Species=>   'Species', #Set only if this SpeciesGraph is bound to a particular Species
  MatchOnce=> '$',   # Map this pattern at most 1 time to a given species
  Fixed=> '$',	     # Concentration of species referred to by this pattern is constant.
  IsCanonical=> '$'
};

my $SpeciesLabel="Auto"; # Allowed values are Auto, HNauty, Quasi

sub setSpeciesLabel{
  my $label=shift;
  my %valid=('Auto'=>1,'HNauty'=>1,'Quasi'=>1);
  if (defined($valid{$label})){
    $SpeciesLabel= $label;
    print "SpeciesLabel method set to $label.\n";
  } else {
    return("Invalid value for SpeciesLabel function: $label");
  }
  return("");
}

sub readString{
  my $sg=shift;
  my $strptr= shift;
  my $clist= shift;
  my $is_species= (@_) ? shift : 1;
  my $stop= (@_) ? shift : "";
  my $mtlist= (@_) ? shift : "";
  my $AllowNewTypes= (@_) ? shift : 0;

  my $string_left=$$strptr;
  my %mlabels=();
  my %elabels=();
  *molecules= $sg->Molecules;

  # Delete any whitespace
#  $string_left=~ s/\s+//g; # Remove whitespace

  # Header (all characters up to :)
  if ($string_left=~ s/^\s*([^:].*)[:]\s*//){
    my $head= $1;
    # Optional species name
    if ($head=~ s/^([A-Za-z0-9_*\s]+)//){
      $sg->Name($1);
    }
    # Optional species compartment
    if ($head=~ s/^\@([A-Za-z0-9_]+)//){
      my $comp= $clist->lookup($1);
      if ($comp){
	$sg->Compartment($comp);
      } else {
	return("Undefined compartment $1");
      }
    }
    if ($head){
      return("Improper syntax for SpeciesGraph header at $head");
    }
  }

  while($string_left){
    # Handle continuation and stopping
    # Continue characters
    if ($string_left=~ s/^[.]//){
      next;
    }
    elsif ($stop){
      # Stop characters
      if ($string_left=~ /$stop/){
        #printf "str_left=%s\n", $string_left;
	last;
      }
    } 
    
    # Handle modifier syntax
    # Set Fixed attribute by prepending '$'
    if ($string_left=~ s/^\$//){
      $sg->Fixed(1);
    }
    # Quantifier
    elsif ($string_left=~ s/^(=|==|<|<=|>|>=)(\d+)//){
      #printf "Quantifier for %s is %s\n", $sg->toString(), $1.$2;
      my $op= $1;
      if ($op eq '='){ $op= '=='};
      $sg->Quantifier($op.$2);
      if ($is_species){
         return("Quantifier not allowed in species");
      }
    }
    # Attributes listed in {}
    elsif ($string_left=~ s/^[{]//){
      while (!($string_left=~ s/^\}//)){
	my $attr="";
	my $value="";

	# Get attribute name
	if ($string_left=~ s/^([^,=\}]+)//){
	  $attr=$1;
	} else {
	  return("Null attribute for SpeciesGraph at $string_left");
	}
	# Get (optional) attribute value
	if ($string_left=~ s/^=([^,\}]+)//){
	  $value=$1;
	} 
	# Remove trailing comma
	$string_left=~ s/^,//;

	if ($attr eq "MatchOnce"){
	  my $val= booleanToInt($value);
	  if ($val==-1){
	    return("Invalid value $value assigned to Boolean attribute $attr");
	  }
	  $sg->MatchOnce($val);
	} elsif ($attr eq "Fixed"){
	  my $val= booleanToInt($value);
	  if ($val==-1){
	    return("Invalid value $value assigned to Boolean attribute $attr");
	  }
	  $sg->Fixed($val);
	} else {
	  return("Invalid attribute $attr for SpeciesGraph");
	}
      }
    }
    else {
      # Read molecule
      my ($mol, $err)= Molecule::newMolecule(\$string_left, $clist);
      if ($err){ return($err);}
      push @molecules, $mol;
    }
  }
  # Done processing input string

  # Check edge labels for correct definition

  # Sorts molecules and components and contructs edges and adjacency matrix.
  if ($is_species){
    #printf "before sort: %s\n", $sg->toString();
    if (my $err= $sg->sortLabel()){
      return($err);
    }
    #printf "after  sort: %s\n", $sg->toString();

    # Check for correct number of subgraphs
    my ($nsub)=$sg->findConnected2();
    if ($nsub != 1){
      #print "nsub=$nsub\n";
      return( "Species $$strptr is not connected");
    }
  } else {
    $sg->updateEdges();
  }

  # Check that molecules match declared types
  if ($mtlist){
    $err= $mtlist->checkSpeciesGraph($sg,{IsSpecies=>$is_species,AllowNewTypes=>$AllowNewTypes});
    if ($err){return($err);}
  }

  $$strptr= $string_left;
  return("");
}

# Convert a pointer to a SpeciesGraph component or molecule to a label (used in XML output)
sub p_to_label{
  my $sg=shift;
  my $p= shift;
  my $string=shift;

  my @inds= split('\.',$p);
  
  if ($#inds==0){
    $string.= sprintf "_M%d", ($inds[0]+1);
  }
  elsif ($#inds==1){
    $string.= sprintf "_M%d_C%d", ($inds[0]+1), $inds[1]+1;
  }
  else{
    die "Error in p_to_label";
  }
		      
  return($string);
}
  

sub labelHNauty{
  use strict;
  use Data::Dumper;
  my $sg= shift;

  # Construct adjacency matrix and partition
  my %adj= ();
  my @partition=();
  # Contains partitions for molecules
  my %mtypes=();
  # Contains partitions for components
  my %ctypes=();
  # Hash to convert pointers to node index
  my %pointer_index=();
  # Bonds hashed by name
  my %bonds=();

  my $n_mol= scalar(@{$sg->Molecules});
  my $imol=0;
  my $icomp= $n_mol;

  # Loop over molecules and components to 
  # (1) determine pointer to index conversions
  # (2) partition by Molecule and Componoent names (could add connectivity later, if it improves performance)
  # (3) create edges from Molecules to its contained Components
  # (4) collect bonds in a hash (precludes need to call update edges beforehand).
  for my $mol (@{$sg->Molecules}){
    my $mname=$mol->Name.".".$mol->State.".".$mol->Compartment;
    $pointer_index{$imol}= $imol;
    push @{$mtypes{$mname}}, $imol;
    for my $edge (@{$mol->Edges}){
      push @{$bonds{$edge}}, $imol;
    }
    my $jcomp=0;
    for my $comp (@{$mol->Components}){
      my $cname=$mname.".".$comp->Name.".".$comp->State.".".$comp->Compartment;
      my $p= "$imol.$jcomp";
      $pointer_index{$p}= $icomp;
      push @{$ctypes{$cname}}, $icomp;
      $adj{$imol}{$icomp}=[0];
      $adj{$icomp}{$imol}=[0];
      for my $edge (@{$comp->Edges}){
	push @{$bonds{$edge}}, $p;
      }
      ++$icomp;
      ++$jcomp;
    }
    ++$imol;
  }

  # Loop over Edges to determine adjacency matrix for bonds
  for my $ename (keys %bonds){
    my $earr= $bonds{$ename};
    my $istart= $pointer_index{$$earr[0]};
    my $iend= $pointer_index{$$earr[1]};
    $adj{$istart}{$iend}=[1];
    $adj{$iend}{$istart}=[1];
  }

  #print Dumper(\%adj),"\n";

  # Set up partitions; sort order molecules then components.
  for my $type (sort keys %mtypes){
    push @partition, $mtypes{$type};
    #print " | ",join(' ',@{$mtypes{$type}});
  }
  for my $type (sort keys %ctypes){
    push @partition, $ctypes{$type};
    #print " | ",join(' ',@{$ctypes{$type}});
  }
  #print "\n";
#  print "molecules : ", join(' ',sort {$a cmp $b} keys %mtypes),"  ";
#  print "components: ", join(' ',sort {$a cmp $b} keys %ctypes),"\n";

  # Call HNauty to obtain canonical ordering
  (my $perm)= HNauty(\%adj, \%adj, \@partition);

  #print Dumper($perm),"\n";

  # Reorder Components of each Molecule according to canonical order
  my $icomp_start=$n_mol;
  for my $mol (@{$sg->Molecules}){
    my $icomp_end= $icomp_start + $#{$mol->Components};
    my @perm_c= map {$perm->{$_}} ($icomp_start..$icomp_end);
    #print join(' ',@perm_c),"\n";
    my @comp_perm= map {[$perm_c[$_],$mol->Components->[$_]]} (0..$#{$mol->Components});
    @comp_perm= sort {$a->[0]<=>$b->[0]} @comp_perm;
    $mol->Components([map {$_->[1]} @comp_perm]);
    $icomp_start= $icomp_end+1;
  }
  # Reorder Molecules according to canonical order
  # This is probably inefficient since we have the exact mapping
  my @mol_perm= map {[$perm->{$_},$sg->Molecules->[$_]]} (0..$#{$sg->Molecules});
  #map {print $_->[0];} @mol_perm;  print "\n";
  @mol_perm= sort {$a->[0]<=>$b->[0]} @mol_perm;
  $sg->Molecules([map {$_->[1]} @mol_perm]);

  $sg->updateEdges();

  # Create exact string representation
  my $string= $sg->toString(0,$sg->Compartment);
  # Remove attributes from label
  $string=~s/^\$//;       
  $sg->StringExact($string);
  $sg->StringID($string);
  $sg->IsCanonical(1);

  return("");
}

sub labelQuasi{
  my $sg= shift;
  my $isCanonical= (@_) ? shift : 0;

  *molecules= $sg->Molecules;
  *edges= $sg->Edges;
  *adjacency= $sg->Adjacency;

  # Sort components of each molecule
  for my $mol (@molecules){
    *components= $mol->Components;
    @components= sort by_component @components;
  }

  # Sort molecules 
  @molecules= sort by_molecule @molecules;
  
  $sg->updateEdges();

  # Create quasi-canonical string representation
  my $string=$sg->toString(1,$sg->Compartment);
  $string=~s/^{.*}//;       
  $sg->StringID($string);

  # Create exact string representation
  $string=$sg->toString(0,$sg->Compartment);
  $string=~s/^\$//;       
  $sg->StringExact($string);

  $sg->IsCanonical($isCanonical);

  return("");
}

# Modified version using different labeling functions
sub sortLabel{
  my $sg= shift;

  # Determine and assign Compartment 
  if (my $err=$sg->assignCompartment){return($err);}

  if ($SpeciesLabel eq "Auto"){
    return($sg->labelQuasi(0));
  } 
  elsif ($SpeciesLabel eq "Quasi"){
    return($sg->labelQuasi(1)); # Equivalent to setting check_iso=>0
  }
  elsif ($SpeciesLabel eq "HNauty"){
    return($sg->labelHNauty());
  }
}

# assigns sg to a Compartment.  If Compartment is set, checks consistency of
# Molecule Compartments with assignment, returning error if not.  If not set,
# determines correct compartment from Molecule Compartments.  Molecules
# unassigned to a Compartment are assigned the sg Compartment.

sub assignCompartment{
  my $sg=shift;

  my $newcomp;
  my %comps;
  %comps=();
  my $comp_surf="";
  my $comp_sav="";

  # Put Molecule Compartments in a hash
  if (defined($sg->Compartment)){
    my $comp= $sg->Compartment;
    $comps{$comp->Name}=$comp;
    if ($comp->SpatialDimensions==2){
      $comp_surf=$comp;
    }
  }
  for my $mol (@{$sg->Molecules}){
    if (defined($mol->Compartment)){
      my $comp= $mol->Compartment;
      $comps{$comp->Name}=$comp; 
      $comp_sav=$comp;
      # Allow only one Compartment to be a Surface
      if ($comp->SpatialDimensions==2){
	if ($comp_surf && ($comp_surf!=$comp)){
	  my $err= sprintf "Molecule Compartments of %s include more than 1 Surface.", $sg->toString();
	  return($err);
	}
	$comp_surf=$comp;
      }
    }
  }
  $ncomp= scalar(keys %comps);

  # Done if no Molecule Compartments found and no sg Compartment specified
  if ($ncomp==0){
    return("");
  } 
  elsif($ncomp==1){
    my @c= keys %comps;
    $newcomp= $comps{$c[0]};
    #print "newcomp=",$newcomp->Name,"\n";
  }
  else {
    # Allow multiple Compartments only if one is Surface
    if (!$comp_surf){
      return("In multi-Compartment SpeciesGraph no Surface Compartments found");
    }
    # Require other Compartments to be connected to surface
    for my $comp_name (keys %comps){
      next if ($comp_name eq $comp_surf->Name);
      my $comp= $comps{$comp_name};
      if (!($comp_surf->adjacent($comp))){
	return("In multi-Compartment SpeciesGraph Compartment $comp_name is not adajacent to Surface");
      }
    }
    # Set sg Compartment to Surface
    $newcomp=$comp_surf;
  }

  # Check that new Compartment assignment is Compatible with previous
  if (defined($sg->Compartment)){
    if ($newcomp != $sg->Compartment){
      my $err= sprintf "Compartment previously assigned SpeciesGraph (%s) is not compatible with new assignement (%s)", $sg->Compartment->Name, $newcomp->Name;
      return($err);
    }
  }

  # Assign Compartment
  $sg->Compartment($newcomp);

  # Set Compartment of Molecules with no Compartment defined to sg Compartment
  for my $mol (@{$sg->Molecules}){
    if (!($mol->Compartment)){
      $mol->Compartment($newcomp);
    }
  }
  return("");
}

# Can be used to create a copy of a single graph or combine multiple graphs into a single graph
sub copy{
  my @sgs= @_;
  my $sg_copy= SpeciesGraph->new();
  
  my $n_edges=0;
  my @msnew=();
  my $isg=1;
  for my $sg (@sgs){
    *molecules= $sg->Molecules;
    for my $imol (0..$#molecules){
      # This form of Molecule->copy adds a prefix to the edges to avoid
      # overlap of edge names between different SpeciesGraphs
      push @msnew, $molecules[$imol]->copy("${isg}_");
    }
    ++$isg;
  } # END loop over SpeciesGraphs
  $sg_copy->Molecules([@msnew]);
  $sg_copy->updateEdges();
  return($sg_copy);
}

sub copySubgraph{
  my $sg=shift;
  my $subgraph= shift;
  my $nsub= shift;
  my $sg_copy= SpeciesGraph->new();
  
  my $n_edges=0;
  my @msnew=();
  *molecules= $sg->Molecules;
  my %enames=();
  for my $imol (0..$#molecules){
    next unless ($$subgraph[$imol]==$nsub);
    push @msnew, $molecules[$imol]->copy();
  }
  $sg_copy->Molecules([@msnew]);
  $sg_copy->updateEdges();
  return($sg_copy);
}

sub stoich{
  my $sg= shift;
  my $mname= shift;
  my $count=0;

  for my $mol (@{$sg->Molecules}){
    next if ($mname cmp $mol->Name);
    ++$count;
  }
  return($count);
}

sub updateEdges{
  my $sg= shift;
  my $AllowDangling= (@_) ? shift : 1;

  *molecules= $sg->Molecules;
  *edges= $sg->Edges;
  *adjacency= $sg->Adjacency;

  # Update edges
  # Loop over molecules and components again to set pointers based
  # on sorted positions
  my %ledges=();
  for my $imol (0..$#molecules){
    $mol= $molecules[$imol];
    *components= $mol->Components;
    for $icomp (0..$#components){
      # Set pointers for edge entries of each component 
      *elabels= $components[$icomp]->Edges;
      my @wildcard=();
      if (@elabels){
	while (@elabels){
	  my $elabel= shift(@elabels);
	  if ($elabel=~/^[*+?]$/){
	    push @wildcard, $elabel;
	  } else {
	    push @{$ledges{$elabel}}, [($imol,$icomp)];
	  }
	}
	# This information will replaced when adjacency hash is updated
	if (@wildcard){
	  $components[$icomp]->Edges([@wildcard]);
	}
      }
    } # END loop over components
  } # END loop over molecules

  @edges=();
  for $elabel (sort keys %ledges){
    my $p1,$p2;
    *parray= $ledges{$elabel};
    $p1="$parray[0][0].$parray[0][1]";
    if ($#parray==1){
      $p2="$parray[1][0].$parray[1][1]";
      push @edges, "$p1 $p2";
    } elsif ($AllowDangling){ # Allow unterminated edges in pattern graphs
      push @edges, "$p1";
    }
  }
  @edges= sort by_edge @edges;
  %adjacency=();
  for my $iedge (0..$#edges){
    my ($p1, $p2)= split(' ',$edges[$iedge]);
    my $elabel= $iedge;
    my ($m1, $c1)= split('\.',$p1);
    push @{$molecules[$m1]->Components->[$c1]->Edges}, $elabel;
    if ($p2){
      $adjacency{$p1}{$p2}= $elabel;
      $adjacency{$p2}{$p1}= $elabel;
      my ($m2, $c2)= split('\.',$p2);
      push @{$molecules[$m2]->Components->[$c2]->Edges}, $elabel;
    } else {
      $adjacency{$p1}= $elabel;
    }
  }
  return($sg);
}


sub toString{
  my $sg= shift;
  my $suppress_edge_names= (@_) ? shift : 0;
  my $speciesCompartment= (@_) ? shift : "";
  my $string="";

  # header
  # NOTE: printing name messes up use of StringExact for hashing species
  #if ($sg->Name){
  #$string.=$sg->Name; 
  #}
  if ($sg->Compartment){
    $string.= '@'.$sg->Compartment->Name;
  }
  if ($string){
    $string.=":";
  }

  # attributes
  if (!$suppress_edge_names){
    my @attr=();
    if ($sg->MatchOnce){
      push @attr, "MatchOnce";
    }
    if (@attr){
      $string.= '{'.join(',',@attr).'}';
    }
    # Handle fixed by prepending '$'
    if ($sg->Fixed){
      $string.= '$';
    }
  }

  $imol=0;
  for my $mol (@{$sg->Molecules}){
    if($imol){
      $string.=".";
    }
    $string.= $mol->toString($suppress_edge_names,$speciesCompartment);
    ++$imol;
  }

  $string.= $sg->Quantifier;
  return($string);
}

# NOTE: Doesn't return enclosing contained because it is assumed that this is called from specific 
#       class containing a SpeciesGraph, e.g. Species, or Reactants or Products in RxnRule.

sub toXML{
  my $sg= shift;
  my $indent= shift;
  my $type=shift;
  my $id=shift;
  my $attributes=shift;

  my $string=$indent."<$type";

  # Attributes
  # id
  $string.= " id=\"".$id."\"";

  # other attributes
  if ($attributes ne ""){$string.= " ".$attributes;}

  # Compartment
  if ($sg->Compartment){
    $string.= " compartment=\"".$sg->Compartment->Name."\"";
  }

  # Objects contained
  my $indent2= "  ".$indent;
  my $ostring="";

  # Molecules
  if (@{$sg->Molecules}){
    $ostring.=$indent2."<ListOfMolecules>\n";
    my $index=1;
    for my $mol (@{$sg->Molecules}){
      $ostring.= $mol->toXML("  ".$indent2,$id,$index);
      ++$index;
    }
    $ostring.=$indent2."</ListOfMolecules>\n";
  }
  # Bonds
  if (@{$sg->Edges}){
    my $bstring="";
    my $index=1;
    my $indent3= "  ".$indent2;
    for my $edge (@{$sg->Edges}){
      my ($p1, $p2) = split(' ',$edge);
      next unless $p2 !=""; # Only print full bonds; half-bonds handled by BindingState variable in Components
      my $bid= sprintf "${id}_B%d", $index;
      $bstring.= $indent3."<Bond";
      $bstring.= " id=\"".$bid."\"";
      $bstring.= " site1=\"".$sg->p_to_label($p1,$id)."\"";
      $bstring.= " site2=\"".$sg->p_to_label($p2,$id)."\"";
      $bstring.= "/>\n";
      ++$index;
    }
    if ($bstring){
      $ostring.= $indent2."<ListOfBonds>\n".$bstring.$indent2."</ListOfBonds>\n";
    }
  }
  
  # Termination
  if ($ostring){
    $string.=">\n"; # terminate tag opening
    $string.= $ostring;
    $string.=$indent."</$type>\n";
  } else {
    $string.="/>\n"; # short tag termination
  }

  return($string);
}

sub addEdge{
  my $sg= shift;
  my $ename= shift;

  *molecules= $sg->Molecules;

  for my $p (@_){
    my ($im, $ic)= split('\.', $p);
    push @{$molecules[$im]->Components->[$ic]->Edges}, $ename;
  }

  return $sg;
}

sub deleteEdge{
  my $sg= shift;
  my $p1= shift;
  my $p2= shift;

  *molecules= $sg->Molecules;
  *adj= $sg->Adjacency;

  return unless defined($adjacency{$p1}{$p2});
  my $ename= $adjacency{$p1}{$p2};

  my $ndrop=0;
  for my $p ($p1,$p2){
    my ($im, $ic)= split('\.', $p);
    *cedges= $molecules[$im]->Components->[$ic]->Edges;
    for my $ie (0..$#cedges){
      if ($cedges[$ie] eq $ename){
	splice @cedges, $ie, 1;
	++$ndrop;
	last;
      }
    }
  }

  if ($ndrop!=2){
    exit_error("deleteEdge acted $ndrop times instead of the correct 2");
  }

  return $sg;
}

# Do breadth first search to split graph into connected subgraphs. Returns
# array of arrays containing indices of molecules in distinct connected
# subgraphs contained in the graph.
sub findConnected{
  my $sg= shift;

  *molecules= $sg->Molecules;
  *edges= $sg->Edges;
  *adjacency= $sg->Adjacency;
  
  my @reached_total= (0)x@molecules;
  my @subgraphs=();
  while(1){
    # Find first molecule that hasn't been included in a subgraph
    my $istart=-1;
    for my $i (0..$#reached_total){
      if ($reached_total[$i]==0){
	$istart=$i;
	last;
      }
    }
    last if ($istart<0);
    
    # Start search from first unreached molecule
    my @reached=       (0)x@molecules;
    $reached[$istart]=1;
    my @mol_new=($istart);
#    print "istart=$istart\n";
    while(@mol_new){
      my @mol_next=();
      # Loop over molecules in mol_new
      for my $imol (@mol_new){
	my $mol= $molecules[$imol];
	# Loop over components in molecule
	for my $icomp (0..$#{$mol->Components}){
	  my $p="$imol.$icomp";
	  # Loop over edges from component
	  for my $q (keys %{$adjacency{$p}}){
	    my ($im, $ic)= split('\.',$q);
	    if (!$reached[$im]){
	      $reached[$im]=1;
	      push @mol_next, $im;
	    }
	  }
	}
      }
      @mol_new=@mol_next;
    }
    for my $i (0..$#reached){
      $reached_total[$i]|= $reached[$i];
    }

    push @subgraphs, [@reached];
  }
  return(@subgraphs);
}

# Modified version that returns n_subgraphs 
# and single array containing index of subgraph
# to which molecule with each index corresponds
sub findConnected2{
  my $sg= shift;

  *molecules= $sg->Molecules;
  *edges= $sg->Edges;
  *adjacency= $sg->Adjacency;
  
  my @subgraph= (0)x@molecules;
  my $n_subgraph=0;
  while(1){
    # Find first molecule that hasn't been included in a subgraph
    my $imol_start=-1;
    for my $imol (0..$#subgraph){
      if ($subgraph[$imol]==0){
	$imol_start=$imol;
	last;
      }
    }
    last if ($imol_start<0);
    ++$n_subgraph;
    
    # Start search from first unreached molecule
    $subgraph[$imol_start]=$n_subgraph;
    my @mol_new=($imol_start);
    while(@mol_new){
      my @mol_next=();
      # Loop over molecules in mol_new
      for my $imol (@mol_new){
	my $mol= $molecules[$imol];
	# Loop over components in molecule
	for my $icomp (0..$#{$mol->Components}){
	  my $p="$imol.$icomp";
	  # Loop over edges from component
	  for my $q (keys %{$adjacency{$p}}){
	    my ($im, $ic)= split('\.',$q);
	    if (!$subgraph[$im]){
	      $subgraph[$im]=$n_subgraph;
	      push @mol_next, $im;
	    }
	  }
	}
      }
      @mol_new=@mol_next;
    }
    #print join(" ",@subgraph),"\n";
  }
  return($n_subgraph, [@subgraph]);
}

# Determine if SpeciesGraph is isomorphic to another SpeciesGraph
# Assume that molecules have already been sorted by molecule name and component state
# using cmp_molecule and cmp_component and cmp_edge
sub isomorphicTo{
  my $sg1= shift;
  my $sg2= shift;

  if ($sg1->StringID cmp $sg2->StringID){
    #print "Failed string test\n";
    return(0);
  }

  # Check for exact match 
  if (!($sg1->StringExact cmp $sg2->StringExact)){
#    print "Exact match\n";
    return(1);
  } 
  # If string is canonical, we're done
#  else {
#    return(0);
#  }

#  printf "Doing isomorphism check...";

  # Nested depth first search, first molecules, then components to find 
  # match
  *molecules1= $sg1->Molecules;
  *molecules2= $sg2->Molecules;

  my @maps=();
  my $nmol= $#molecules1;
  my @mused=(0) x @molecules1;
  my @mptr= (0) x @molecules1;
  my $im1=0;
  my @cptrs=();
  my @cuseds=();
  *components1= $molecules1[$im1]->Components;
  my $im2, $ic1, $ic2, $ncomp=$#components1;

  *adj1= $sg1->Adjacency;
  *adj2= $sg2->Adjacency;
 
  # depth first search over Molecules
 MITER:
  while(1){
    # find a match at the current level
    # Currently loop is done over all possible molecules, but this could be
    # changed to loop over molecules adjacent to molecules higher level to 
    # limit search.  
    my $mmatch=0;
#   for $im2 ($mptr[$im1]..$nmol){
    for ($im2=$mptr[$im1]; $im2<=$nmol; ++$im2){
      next if $mused[$im2]; # Continue if this molecule already mapped
      next if ($molecules1[$im1]->Name cmp $molecules2[$im2]->Name);
      next if ($molecules1[$im1]->Compartment != $molecules2[$im2]->Compartment);
      # Initialize data for component match at this level
      $mptr[$im1]= $im2;
      *components2= $molecules2[$im2]->Components;
      $cptrs[$im1]= [(0) x @components1];
      $cuseds[$im1]= [(0) x @components2];
      $ic1=0;
      $mmatch=1;
      last;
    }

    # Move up a level (to last component of molecule at previous level)
    # if no match molecules found
    if (!$mmatch){
      last MITER if ($im1==0);
      # Reset molecule pointer at current level
      $mptr[$im1]=0;
      --$im1;
      *components1= $molecules1[$im1]->Components;
      $ncomp= $#components1;
      $ic1= $ncomp;
      $im2= $mptr[$im1];
      ++$cptrs[$im1][$ic1];
      $mused[$mptr[$im1]]=0;
      *components2= $molecules2[$im2]->Components;
#      print "set im1 to $im1  $cptrs[$im1][$ic1]\n";
    }
    
    # Do depth first search over components of molecule 2
    *cptr= $cptrs[$im1];
    *cused= $cuseds[$im1];
  CITER:
    while(1){
      $cmatch=0;
      for ($ic2=$cptr[$ic1]; $ic2<=$ncomp;++$ic2){
	next if $cused[$ic2];
	if ($components1[$ic1]->compare_local($components2[$ic2])){
	  next;
	}
	
	#Check component edges
	my $ematch=1;
	$p1="$im1.$ic1";
	$p2="$im2.$ic2";
      EDGE:
	for my $q1 (keys %{$adj1{$p1}}){
	  my ($jm1, $jc1)= split('\.',$q1);
	  next if ($jm1>$im1);
	  if ($jm1==$im1){
	    next if ($jc1>=$ic1);
	    #	  exit_error("isomorphicTo can't handle bonds among components of same molecule");
	  }
	  my $q2 = "$mptr[$jm1].$cptrs[$jm1][$jc1]";
	  if (!defined( $adj2{$p2}{$q2})){
	    #	  print "Failed adjacency: $p1 $q1 $p2 $q2\n";
	    $ematch=0;
	    last EDGE;
	  }
	}
	next unless ($ematch);
	
	$cptr[$ic1]= $ic2;
	# Complete mapping of this molecule if $ic1==$ncomp
	if ($ic1==$ncomp){
	  $cmatch=1;
	  last;
	} else {
	  # descend to next component
	  $cused[$ic2]=1;
	  ++$ic1;
	  #	print "ic1=$ic1 cptr=$cptr[$ic1] ncomp=$ncomp\n";
	  next CITER;
	}
      }
      
      # Move up a component level if no match found
      if (!$cmatch){
	# Move to next molecule at current level if up exhausted
	# component search
	if ($ic1==0){
	  # Increment molecule pointer at current level
	  ++$mptr[$im1];
	  #	print "set im2 to $mptr[$im1] ", join(' ',@mused), "\n";
	  next MITER;
	}
	# Reset component pointer at current level
	$cptr[$ic1]=0;
	--$ic1;
	$cused[$cptr[$ic1]]=0; # Reset pointers at new level
	++$cptr[$ic1];
	#      print "set ic1 to $ic1 cptr is $cptr[$ic1]\n";
	next CITER;
      }
      last CITER;
    } # END CITER

    # If $im1==$nmol, then graphs are isomorhpic and we can return
    if ($im1==$nmol){
      return(1); 
    } 

    # Move down a level in molecules (increment $im1) 
    $mused[$im2]=1;
    ++$im1;
    *components1= $molecules1[$im1]->Components;
    $ncomp= $#components1;
  } # END MITER

  #print "Failed full isomorphism test\n";
  return(0);
}

# Determine if SpeciesGraph is isomorphic 
# to a portion of another SpeciesGraph
sub isomorphicToSubgraph{
  my $sg1= shift;
  my $MatchOnce= $sg1->MatchOnce;

  *molecules1= $sg1->Molecules;
  *edges1= $sg1->Edges;
  *adj1= $sg1->Adjacency;
  my @maps=();

 GRAPH: while (my $sg2= shift){
    *molecules2= $sg2->Molecules;
    # Number of molecules
    if (scalar(@molecules1) > scalar(@molecules2)){
      next;
    }
    
    # Check that number of edges is same
    *edges2= $sg2->Edges;
    if (scalar(@edges1) > scalar(@edges2)){
      next;
    }
    
    # Nested depth first search, first molecules, then components to find 
    # match
    
    my $nmol1= $#molecules1;
    my $nmol2= $#molecules2;
    my @mptr= (0) x @molecules1;
    my @mused=(0) x @molecules2;
    my $im1=0;
    my @cptrs=();
    my @cuseds=();
    *components1= $molecules1[$im1]->Components;
    my $im2, $ic1, $ic2, $ncomp1=$#components1;
    my $ci1, $ci2;
    
    *adj2= $sg2->Adjacency;
    
    # depth first search over Molecules
  MITER:
    while(1){
      # find a match at the current level
      # Currently loop is done over all possible molecules, but this could be
      # changed to loop over molecules adjacent to molecules higher level to 
      # limit search.  
      my $mmatch=0;
      for ($im2=$mptr[$im1]; $im2<=$nmol2; ++$im2){
	next if $mused[$im2]; # Continue if this molecule already mapped
	my $namestring= $molecules1[$im1]->Name;
	$namestring=~ s/\*$/.*/;
	#print "namestring=$namestring\n";
	next unless ($molecules2[$im2]->Name=~ /^${namestring}$/);
	#next if ($molecules1[$im1]->Name cmp $molecules2[$im2]->Name);
	if ($molecules1[$im1]->Compartment){
	  next if ($molecules1[$im1]->Compartment != $molecules2[$im2]->Compartment);
	}
	#print "$im1 $im2: $molecules1[$im1]->Name -> $molecules2[$im2]->Name\n";;
	# Initialize data for component match at this level
	$mptr[$im1]= $im2;
	*components2= $molecules2[$im2]->Components;
	$cptrs[$im1]= [(0) x @components1];
	$cuseds[$im1]= [(0) x @components2];
	$ic1= ($ncomp1>=0) ? 0 : -1;
	$mmatch=1;
	last;
      }
      
      # Move up a level (to last component of molecule at previous level)
      # if no match molecules found
      if (!$mmatch){
	last MITER if ($im1==0);
	# Reset molecule pointer at current level
	$mptr[$im1]=0;
	--$im1;
	*components1= $molecules1[$im1]->Components;
	$ncomp1= $#components1;
	$ic1= $ncomp1;
	$im2= $mptr[$im1];
	$mused[$mptr[$im1]]=0;
	if ($ic1>=0){
	  ++$cptrs[$im1][$ic1];
	} else {
	  ++$mptr[$im1];
	  next MITER;
	}
	*components2= $molecules2[$im2]->Components;
	#print "set im1 to $im1  $cptrs[$im1][$ic1]\n";
      }
      
      # Do depth first search over components of molecule 2
      *cptr= $cptrs[$im1];
      *cused= $cuseds[$im1];
    CITER:
      while(1){
	if ($ncomp1>=0){
	  $ci1= $components1[$ic1];
	  $ncomp2= $#components2;
	  $cmatch= 0;
	  for ($ic2=$cptr[$ic1]; $ic2<=$ncomp2;++$ic2){
	    #      print "ic2=$ic2\n";
	    next if $cused[$ic2];
	    $ci2= $components2[$ic2];
	    # Component name
	    next if ($ci1->Name cmp $ci2->Name);
	    # Component state only if present in sg1
	    if (defined($ci1->State)){
	      if ($ci1->State=~ /[*+?]/){
		if ($ci1->State eq '+'){
		  next if ($ci2->State eq '');
		}
	      } else {
		next if ($ci1->State cmp $ci2->State);
	      }
	    }
	    if ($ci1->Compartment){
	      next if ($ci1->Compartment != $ci2->Compartment);
	    }
	    #      print " $ic1 $ic2 $ci1->Name -> $ci2->Name\n";
	    
	    # Number of component edges must match (primarily used to look for free
	    # binding sites
	    # Number of edges
	    my $diff= $#{$ci2->Edges}-$#{$ci1->Edges};
	    if ($diff){
	      # Mismatch unless first Edge is wildcard
	      my $wild= $ci1->Edges->[0];
	      next unless ($wild =~ /^[*+?]$/);
	      # + wildcard requires $diff>=0 (= case handled above)
	      if ($wild eq '+'){
		next unless ($diff>0);
	      } else {
		# *? (equivalent) wildcard requires $diff>-1, #c2 edges >= #c1 edges - 1 (for wildcard)
		next unless ($diff>=-1);
	      }
	    }
	    
	    #Check component edges
	    my $ematch=1;
	    $p1="$im1.$ic1";
	    $p2="$im2.$ic2";
	  EDGE:
	    for my $q1 (keys %{$adj1{$p1}}){
	      my ($jm1, $jc1)= split('\.',$q1);
	      next if ($jm1>$im1);
	      if ($jm1==$im1){
		next if ($jc1>=$ic1);
	      }
	      my $q2 = "$mptr[$jm1].$cptrs[$jm1][$jc1]";
	      if (!defined( $adj2{$p2}{$q2})){
		#print "Failed adjacency: $p1 $q1 $p2 $q2\n";
		$ematch=0;
		last EDGE;
	      }
	    }
	    next unless ($ematch);
	    
	    $cptr[$ic1]= $ic2;
	    # Complete mapping of this molecule if $ic1==$ncomp1
	    if ($ic1==$ncomp1){
	      $cmatch=1;
	      last;
	    } else {
	      # descend to next component
	      $cused[$ic2]=1;
	      ++$ic1;
	      #	print "ic1=$ic1 cptr=$cptr[$ic1] ncomp1=$ncomp1\n";
	      next CITER;
	    }
	  }
	} else {
	  # No components in pattern
	  $cmatch=1;
	}
      
	# Move up a component level if no match found
	if (!$cmatch){
	  # Move to next molecule at current level if up exhausted
	  # component search
	  if ($ic1<=0){
	    # Increment molecule pointer at current level
	    ++$mptr[$im1];
	    #print "set im2 to $mptr[$im1] $nmol1 ", join(' ',@mused), "\n";
	    next MITER;
	  }
	  # Reset component pointer at current level
	  $cptr[$ic1]=0;
	  --$ic1;
	  $cused[$cptr[$ic1]]=0; # Reset pointers at new level
	  ++$cptr[$ic1];
	  #      print "set ic1 to $ic1 cptr is $cptr[$ic1]\n";
	  next CITER;
	}
	
	# If $im1==$nmol, then graphs are isomorhpic and we can return
	# Modify to save map for subgraph isomorphism case
	if ($im1==$nmol1){
	  my $map= Map->new;
	  $map->Source($sg1);
	  $map->Target($sg2);
	  my %mapf=();
	  for my $im (0..$nmol1){
	    my $im2= $mptr[$im];
	    $mapf{$im}= $im2;
	    *cptr= $cptrs[$im];
	    #print "$im -> $im2\n";
	    for my $ic (0..$#cptr){
	      $mapf{"$im.$ic"}= "$im2.$cptr[$ic]";
	      #print "$im.$ic -> $im2.$ic2\n";
	    }
	  }
	  $map->MapF({%mapf});
	  push @maps, $map;
	  if ($MatchOnce){
	    next GRAPH;
	  }
	  if ($ncomp1>=0){
	    ++$cptr[$ic1];
	    next CITER;
	  } else {
	    # Go to next molecule if no components in the current molecule
	    #print " ncomp1=$ncomp1 im1=$im1 $mptr[$im1]\n";
	    ++$mptr[$im1];
	    next MITER;
	  }
	} 
	last CITER;
      } # END CITER

      # Move down a level in molecules (increment $im1) 
      $mused[$im2]=1;
      ++$im1;
      *components1= $molecules1[$im1]->Components;
      $ncomp1= $#components1;
      #    print "set im1 to $im1\n";
    } # END MITER
  } # END loop over $sg2  
  return(@maps);
}

# Find mapping of graph onto another graph (or set of graphs) according to the following rules
# For each molecule: 
# 1. Preserve molecule name and label (if defined)
# 2. Preserve component number and names
# 
# In general, there may be multiple mappings.  For reaction rules it is
# necessary to select a single mapping.  This can be done be either taking the
# default mapping to be the one that takes the first possible match

# STILL NEEDS WORK


sub findMaps{
  my $sg1= shift;

  *molecules1= $sg1->Molecules;
  my @maps=();

  while (my $sg2= shift){
    *molecules2= $sg2->Molecules;

    my $nmol1= $#molecules1;
    my $nmol2= $#molecules2;
    my @mused=(0) x @molecules2;
    my @mptrs= (-1) x @molecules1;
    my @cptrs=(0) x @molecules1;
    my $im1, $im2, $ic1, $ic2;
    my $ci1, $ci2;

    # Loop over molecules of Source
    for ($im1=0; $im1<= $nmol1; ++$im1){
      *components1= $molecules1[$im1]->Components;
      $ncomp1=$#components1;
      my $mmatch=0;
    MOL:
      # Loop over molecules of Target
      for ($im2=0; $im2<=$nmol2; ++$im2){
	next if $mused[$im2]; # Continue if this molecule already mapped
	next if ($molecules1[$im1]->Name cmp $molecules2[$im2]->Name);
	next if ($molecules1[$im1]->Label cmp $molecules2[$im2]->Label);
	*components2= $molecules2[$im2]->Components;
	$ncomp2=$#components2;
	next if ($ncomp1!=$ncomp2);
	my @cptr=  (0) x @components1;
	my @cused= (0) x @components2;
	#print "$molecules1[$im1]->Name -> $molecules2[$im2]->Name\n";
      COMP:      
	for ($ic1=0; $ic1<= $ncomp1; ++$ic1){
	  $ci1= $components1[$ic1];
	  my $cmatch=0;
	  for ($ic2=0; $ic2<= $ncomp2; ++$ic2){
	    #printf "%s -> %s\n", $components1[$ic1]->Name, $components2[$ic2]->Name;
	    next if $cused[$ic2];
	    $ci2= $components2[$ic2];
	    # Component name
	    next if ($ci1->Name cmp $ci2->Name);
	    next if ($ci1->Label cmp $ci2->Label);
	    $cmatch=1;
	    $cptr[$ic1]= $ic2;
	    $cused[$ic2]=1;
	    last;
	  } # End loop over $ic2
	  # Unmatched components invalidate match for current molecule $im2
	  if (!$cmatch){
	    next MOL;
	  }
	} # End loop over $ic1
	# There is one-to-one mapping of components of $im1 to components of $im2
	$mmatch=1;
	$mptrs[$im1]= $im2;
	@cptrs[$im1]= [@cptr];
	$mused[$im2]=1;
	last;
      } # End loop over $im2
      # If no match for current molecule could be found, $mptrs[$im1]=-1
    } # End loop over $im1

    # Create new Map object for completed map, which is the best mapping that could be found
    my $map= Map->new;
    $map->Source($sg1);
    $map->Target($sg2);
    my %mapf=();
    for my $im1 (0..$nmol1){
      my $im2= $mptrs[$im1];
      $mapf{$im1}= $im2;
      if ($im2<0){
	# Check for component label mappings
	next;
      }
      *cptr= $cptrs[$im1];
      for my $ic1 (0..$#cptr){
	$mapf{"$im1.$ic1"}= "$im2.$cptr[$ic1]";
      }
    }
    $map->MapF({%mapf});
    push @maps, $map;
  } # End loop over target graphs
  
  return(@maps);
}

# A simplified replacement for findMaps. Finds mapping by first labeling the
# molecules and components and storing a pointer to the location of each
# label. By looping over labels of either the first or the second graph, the
# forward or reverse map can be generated.

sub findMaps2{
  my @ptrs;
  @ptrs=();
  my @sgs= @_;
  for $sg (@sgs){
    *molecules= $sg->Molecules;
    my %ptr, %labels, @clabels;
    %ptr=(); %labels=(); my $im=0;
    for my $mol (@molecules){
      my $mlabel;
      # Get component labels
      @clabels=(); 
      for my $comp (@{$mol->Components}){
	if ($clabel=$comp->Label){
	  push @clabels, '%'.$clabel;
	} else {
	  push @clabels, $comp->Name;
	}
      }
      # User provided label supercedes other labeling
      # It must be unique for each molecule and 
      # each component
      if (!($mlabel=$mol->Label)){
	# Automatic mol label starts with molname
	$mlabel= $mol->Name;
	$mlabel.= '_'.join('_',sort @clabels).'_';
	$mlabel.= ++$labels{$mlabel};
      } else {
	$mlabel= '%'.$mlabel;
      }
      $ptr{$mlabel}= $im;
      # Component labels have syntax mlabel|clabel.  If user label is
      # provided, it has precedence and mlabel is omitted. Otherwise, $clabel
      # is name of component plus a number that indicates its order in
      # occurence of identical components
      for my $ic (0..$#clabels){
	my $clabel= $clabels[$ic];
	if (!($clabel=~ /^\%/)){
	  # Automatic label
	  $clabel= $mlabel.'|'.$clabel.'_';
	  $clabel.= ++$labels{$clabel};
	}
	$ptr{$clabel}= "$im.$ic";
      }
      ++$im;
    }
    push @ptrs, {%ptr};
  }
    
  # Set up forward map
  my $map= Map->new;
  $map->Source($sgs[0]);
  $map->Target($sgs[1]);
  my %mapf, %mapr;
  %mapf=(); %mapr=();
  *ptr1= $ptrs[0];
  *ptr2= $ptrs[1];
  for my $label (keys %ptr1){
    my $p1= $ptr1{$label};
    my $p2= (defined($ptr2{$label})) ? $ptr2{$label} : -1;
    #print "label $label: $p1 $p2\n";
    $mapf{$p1}= $p2;
  }  
  $map->MapF({%mapf});
  # Reverse map
  for my $label (keys %ptr2){
    my $p2= $ptr2{$label};
    my $p1= (defined($ptr1{$label})) ? $ptr1{$label} : -1;
    #print "label $label: $p2 $p1\n";
    $mapr{$p2}= $p1;
  }  
  $map->MapR({%mapr});
  return(($map));
}

# Canonical order for molecules
sub cmp_molecule{
  my $a= shift;
  my $b= shift;
  my $cmp;
  if ($cmp=($a->Name cmp $b->Name)){
    return($cmp);
  }
  *comp_a=$a->Components;
  *comp_b=$b->Components;
  if ($cmp=($#comp_a <=> $#comp_b)){
    return $cmp;
  }

  # Comparison of Compartment Names
  if ($a->Compartment){
    return(1) if !($b->Compartment);
    if ($cmp=($a->Compartment->Name cmp $b->Compartment->Name)){
      return($cmp);
    }
  }
  if ($b->Compartment){
    return(-1) if !($a->Compartment);
  }

  for my $i (0..$#comp_a){
    if ($cmp=cmp_component($comp_a[$i], $comp_b[$i])){
      return $cmp;
    }
  }
  return(0);
}

sub by_molecule{
  return (&cmp_molecule($a,$b));
}

# Canonical order for components of molecules
sub cmp_component{
  my $a= shift;
  my $b= shift;

  my $cmp;

  # Lexical comparison of name
  if ($cmp=($a->Name cmp $b->Name)){
    return($cmp);
  }
  # Lexical comparison of state labels
  if ($cmp=($a->State cmp $b->State)){
    return($cmp);
  }

  # Comparison of number of edges
  *a_edges= $a->Edges;
  *b_edges= $b->Edges;
  if ($cmp=($#b_edges <=> $#a_edges)){
    return($cmp);
  }

  # Comparison of edges
#  for my $i (0..$#a_edges){
#    if ($cmp=($a_edges[$i] cmp $b_edges[$i])){
#      return($cmp);
#    }
#  }

  return(0);
}

sub by_component{
  return (&cmp_component($a,$b));
}

sub cmp_edge{
  my $a= shift;
  my $b= shift;
  my $cmp;

  my (@a_p)= split('[ \.]', $a);
  my (@b_p)= split('[ \.]', $b);
  
  for my $i (0..$#a_p){
    if($cmp=($a_p[$i]<=>$b_p[$i])){
      return($cmp);
    }
  }
  # Getting here means edges are identical, which shouldn't happen.
  return(0);
}

sub by_edge{
  return (&cmp_edge($a,$b));
}

1;
